import struct
import time
from . import Future, Condition, gethub
from collections import deque

class BaseChannel(object):
    def __init__(self):
        self._pending = deque()
        self._cond = Condition()
        hub = gethub()
        hub.do_spawnhelper(self.sender)
        hub.do_spawnhelper(self.receiver)

    def peek_request(self):
        self.wait_requests()
        return self._pending[0]

    def pop_request(self):
        return self._pending.popleft()

    def wait_requests(self):
        while not self._pending:
            self._cond.wait()

    def get_pending_requests(self):
        while self._pending:
            yield self._pending.popleft()

class PipelinedReqChannel(BaseChannel):

    def __init__(self):
        super().__init__()
        self._producing = deque()
        self._cur_producing = []

    def produce(self, value):
        val = self._cur_producing.append(value)
        num = self._producing[0][0]
        if num is None:
            self._producing.popleft()[1].set(value)
        elif len(self._cur_producing) >= num:
            res = tuple(self._cur_producing)
            self._producing.popleft()[1].set(res)
        else:
            return
        del self._cur_producing[:]

    def request(self, input, num_output=None):
        val = Future()
        self._pending.append(input)
        self._producing.append((num_output, val))
        self._cond.notify()
        return val.get()

    def wait_requests(self):
        while True:
            if self._pending:
                return
            self._cond.wait()

    def get_pending_requests(self):
        while self._pending:
            yield self._pending.popleft()

class MuxReqChannel(BaseChannel):
    def __init__(self):
        super().__init__()
        self.requests = {}
        self.init_id()

    def init_id(self):
        import os, random, struct
        self.prefix = struct.pack('HHL',
            os.getpid() % 65536, random.randrange(65536), int(time.time()))
        self.counter = 0

    def new_id(self):
        self.counter += 1
        if self.counter >= 4294967296:
            self.init_id()
            self.counter -= 4294967296
        return self.prefix + struct.pack('L', self.counter)

    def request(self, input):
        id = self.new_id()
        assert not id in self.requests
        val = Future()
        self.requests[id] = val
        self._pending.append((id, input))
        self._cond.notify()
        return val.get()

    def produce(self, id, data):
        self.requests.pop(id).set(data)




import zmq
from zmq import *
import errno
from functools import partial

from . import core, channel

DEFAULT_IO_THREADS = 1

def send_data(sock, *data, address=None):
    if address is None:
        _rep = []
    else:
        _rep = list(address)
    for a in data:
        if isinstance(a, str):
            _rep.append(a.encode('utf-8'))
        else:
            _rep.append(a)
    while True:
        try:
            sock.send_multipart(_rep, zmq.NOBLOCK)
        except zmq.ZMQError as e:
            if e.errno == errno.EAGAIN:
                hub.do_write(sock)
                continue
            elif e.errno == errno.EINTR:
                continue
            else:
                raise
        else:
            break


def rep_responder(sock, address, callback, data):
    hub = core.gethub()
    reply = callback(*data)
    if isinstance(reply, bytes):
        send_data(sock, reply, address=address)
    elif isinstance(reply, str):
        send_data(sock, reply.encode('utf-8'), address=address)
    elif reply is None:
        raise RuntimeError("Replier callback must return either string,"
            " bytes or sequence of strings or bytes")
    else:
        send_data(sock, *reply, address=address)

def rep_listener(sock, callback):
    hub = core.gethub()
    while True:
        hub.do_read(sock)
        try:
            data = sock.recv_multipart(zmq.NOBLOCK)
        except zmq.ZMQError as e:
            if e.errno == errno.EAGAIN or e.errno == errno.EINTR:
                continue
            else:
                raise
        i = data.index(b'')
        addr = data[:i+1]
        data = data[i+1:]
        hub.do_spawn(partial(rep_responder, sock, addr, callback, data))

def rep_socket(callback):
    sock = context().socket(zmq.XREP)
    core.gethub().do_spawnservice(partial(rep_listener, sock, callback))
    return sock

def sub_listener(sock, callback):
    hub = core.gethub()
    while True:
        hub.do_read(sock)
        try:
            data = sock.recv_multipart(zmq.NOBLOCK)
        except zmq.ZMQError as e:
            if e.errno == errno.EAGAIN or e.errno == errno.EINTR:
                continue
            else:
                raise
        hub.do_spawn(partial(callback, *data))

def sub_socket(callback):
    sock = context().socket(zmq.SUB)
    core.gethub().do_spawnservice(partial(sub_listener, sock, callback))
    return sock

def pub_socket():
    return PubChannel()

def req_socket():
    return ReqChannel()

class ReqChannel(channel.MuxReqChannel):

    def __init__(self):
        super().__init__()
        self._sock = context().socket(zmq.XREQ)

    def bind(self, value):
        self._sock.bind(value)

    def connect(self, value):
        self._sock.connect(value)

    def sender(self):
        wait_write = core.gethub().do_write
        while True:
            wait_write(self._sock)
            id, data = self.peek_request()
            self._sock.send(id, zmq.SNDMORE)
            self._sock.send(b"", zmq.SNDMORE)
            self._sock.send_multipart(data)
            self.pop_request()

    def receiver(self):
        wait_read = core.gethub().do_read
        while True:
            wait_read(self._sock)
            data = self._sock.recv_multipart()
            assert data[1] == b''
            self.produce(data[0], data[2:])

class PubChannel(object):

    def __init__(self):
        super().__init__()
        self._sock = context().socket(zmq.PUB)

    def bind(self, value):
        self._sock.bind(value)

    def connect(self, value):
        self._sock.connect(value)

    def publish(self, *args):
        send_data(self._sock, *args)


def _get_fd(value):
    if isinstance(value, zmq.Socket):
        return value
    else:
        return value.fileno()

def plug(hub, io_threads=DEFAULT_IO_THREADS):
    assert not hasattr(hub, 'zmq_context')
    ctx = hub.zmq_context = zmq.Context(io_threads)
    hub.change_poller(zmq.Poller, filedes=_get_fd,
        POLLIN=zmq.POLLIN, POLLOUT=zmq.POLLOUT)
    hub.log_plugged(ctx, name='zmq_context')
    return ctx

def context():
    hub = core.gethub()
    ctx = getattr(hub, 'zmq_context', None)
    if ctx is None:
        return plug(hub)
    return ctx



from .core import Hub, gethub, Future, Condition, Lock
from .net import bufferedsocket
from contextlib import contextmanager
from functools import wraps

__all__ = [
    'Hub',
    'gethub',
    'sleep',
    'timeout',
    'with_timeout',
    'TimeoutException',
    'Future',
    'Condition',
    ]

def sleep(value):
    gethub().do_sleep(value)

class TimeoutException(Exception):
    """Raised when you use timeout context manager and process timed out"""

class FinishedException(Exception):
    """Internal exception for timeout context manager"""

@contextmanager
def timeout(value, description="Process timed out"):
    """Context manager to limit execution time of code path to time in seconds
    """
    def _timeout():
        try:
            sleep(value)
        except FinishedException:
            pass
        else:
            raise TimeoutException(description)
    let = gethub().do_spawn(timeout)
    yield let
    let.throw(FinishedException())

def with_timeout(value, description="Process timed out"):
    """Decorator which limits execution time of function call to time in seconds
    """
    def wrapper(fun):
        @wraps(fun)
        def wrapping(*a, **k):
            with timeout(value, description):
                return fun(*a, **kw)
        return wrapping
    return wrapper



import socket
import errno

"""
This implementation have done under some drug pressure and not tested yet.
But if you want to test some feedback is appreciated.
"""

__all__ = [
    'bufferedsocket',
    ]

class bufferedsocket(object):

    def __init__(self, sock, bufsize=4096):
        self._sock = sock
        self._buf = bytearray()
        self.bufsize = bufsize

    def readline(self, endline='\r\n'):
        while True:
            try:
                idx = self._buf.index(endline)
            except IndexError:
                try:
                    self._read()
                except EOFError:
                    if self._buf:
                        res = self.buf
                        self.buf = bytearray()
                        return res
                    else:
                        raise
            else:
                idx += len(endline)
                res = self.buf[:idx]
                del self.buf[:idx]
                return res

    def read(self, bytes):
        while len(self._buf) < bytes:
            self._read(bytes - len(self._buf))
        res = self.buf[:bytes]
        del self.buf[:bytes]
        return res

    def _read(self, bytes=0):
        gethub().do_read(self._sock)
        try:
            val = self._sock.read(max(bytes, self.bufsize))
        except socket.error as e:
            if e.errno in (EINTR, EAGAIN):
                return
            else:
                raise
        if not val:
            raise EOFError()
        self._buf.extend(val)

    def write(self, bytes):
        hub = gethub()
        while bytes:
            hub.do_write(self._sock)
            try:
                done = self._sock.send(bytes)
            except socket.error as e:
                if e.errno in (EINTR, EAGAIN):
                    continue
                else:
                    raise
            else:
                if not done:
                    raise EOFError()
                bytes = bytes[done:]
